// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: constraint_index.h,v 1.6 2002/09/30 04:51:48 carzanig Exp $
//
#ifndef CONSTRAINT_INDEX_H
#define CONSTRAINT_INDEX_H

#include "fwd_table.h"
#include "ft_allocator.h"

/** generic index of <em>equals</em>, <em>less-than</em>, and
 *  <em>greater-than</em> constraints.
 *
 *  This template index works for <em>equals</em>, <em>less-than</em>,
 *  and <em>greater-than</em> constraints for all the types that
 *  implement operators <code>==</code>, <code>&lt;</code>, and
 *  <code>&gt;</code> operators.  In particular, this template is
 *  optimal for numbers (integers or floating point).
 **/
template<class T>
class constraint_index {
private:
    class node {
    public:
	T value;
	Constraint * c;
	
	node(const T & v, Constraint * d): value(v), c(d) {}

	bool operator < (const T & x) const { return value < x; }
	bool operator > (const T & x) const { return value > x; }
	bool operator == (const T & x) const { return value == x; }
    };

    typedef vector<node> v_index;
    typedef map<T, Constraint *> m_index;

    v_index lt_map;
    v_index gt_map;
    m_index eq_map;
    
public:
    Constraint * add_lt(const T & v, FTAllocator & ftmemory) {
	v_index::iterator i = lower_bound(lt_map.begin(), lt_map.end(), v);
	if (i == lt_map.end() || (*i).value != v)  
	    i = lt_map.insert(i, node(v, new (ftmemory)Constraint()));

	return (*i).c;
    }

    Constraint * add_gt(const T & v, FTAllocator & ftmemory)  {
	v_index::iterator i = lower_bound(gt_map.begin(), gt_map.end(), v);
	if (i == gt_map.end() || (*i).value != v) 
	    i = gt_map.insert(i, node(v, new (ftmemory)Constraint()));

	return (*i).c;
    }

    Constraint * add_eq(const T & v, FTAllocator & ftmemory) {
	pair<m_index::iterator, bool> pib;
	pib = eq_map.insert(m_index::value_type(v, NULL));
	if (pib.second) 
	    (*pib.first).second = new (ftmemory)Constraint();

	return (*pib.first).second;
    }

    bool match(const T & v, CProcessor & p) const {
	//
	// equality first
	//
	m_index::const_iterator i = eq_map.find(v);
	if (i != eq_map.end()) 
	    if (p.process_constraint((*i).second))
		return true;
	//
	// now greater-than
	//
	v_index::const_iterator vi;
	for(vi = gt_map.begin(); vi != gt_map.end() && *vi < v; ++vi) 
	    if (p.process_constraint((*vi).c))
		return true;
	//
	// now less-than
	//
	v_index::const_reverse_iterator vri;
	for(vri = lt_map.rbegin(); vri != lt_map.rend() && *vri > v; ++vri) 
	    if (p.process_constraint((*vri).c))
		return true;
	return false;
    }
};

#endif
