// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: fwd_table.cc,v 1.20 2002/09/30 04:50:25 carzanig Exp $
//
#include "fwdconf.h"

#include <iostream>
#include <algorithm>
#include <list>
#include <set>

#include "siena_types.h"
#include "fwd_table.h"
#include "string_index.h"
#include "bool_index.h"
#include "tst.h"
#include "constraint_index.h"
#include "ft_allocator.h"

#ifndef HAVE_INLINE
#include "fwd_table.icc"
#endif

/** attribute descriptor
 *
 *  this class represents a constraint name within the constraint
 *  index.  It holds all the type-specific indexes for the constraints
 *  pertaining to this attribute.
 **/
class Attribute {
public:
    /** set of interfaces for which this attribute is a
     *  <em>determinant</em> attribute.
     *
     *  Attribute <em>A</em> is <em>determinant</em> for interface
     *  <em>I</em> if all the conjuncts (i.e., filters) contained in
     *  the disjunct (i.e., predicate) associated with <em>I</em>
     *  contain at least one constraint of <em>A</em>.
     **/
    Selectivity *		exclude;

    /** index of <em>integer</em> constraints **/
    constraint_index<int>	int_index;
    /** index of <em>double</em> constraints **/
    constraint_index<double>	double_index;
    /** index of <em>string</em> constraints **/
    StringIndex			str_index;
    /** index of <em>boolean</em> constraints **/
    BoolIndex			bool_index;
    //
    // more constraints can be added here (for example, for Time)
    // ...to be continued...
    //

    /** adds the given constraints to the appropriate constraint
     *	index associated with this attribute name.
     **/
    Constraint *		add_constraint(const sx_constraint &,
					       FTAllocator &);
    /** constructor **/
    Attribute()
	: exclude(NULL), 
	  int_index(), double_index(), str_index(), bool_index() {};
};

/** a selectivity descriptor
 *
 *  associates a <em>selectivity</em> to a given attribute/constraint
 *  name.  The selectivity associated with a name is given by a set of
 *  interfaces that can not be matched by a message unless that
 *  attribute is present in the message.
 **/
class Selectivity {
public:
    /** creates a selectivity item with the given name, excluding the
     *	given interface and connecting it to the given previous
     *  element in the selectivity table. 
     **/
    Selectivity(const char * n, ifid_t i, Selectivity * prev)
	: name(n), exclude(), prev(prev), next(NULL) {}

    /** constraint/attribute name **/
    const char *		name;
    /** set of interfaces excluded by the absence of this attribute **/
    ibitvector			exclude;

    /** pointer to the previous element in the selectivity table.
     *
     *  Points to an element with greater or equal selectivity
     **/
    Selectivity *		prev;	

    /** pointer to the next element in the selectivity table.
     *
     *  Points to an element with less or equal selectivity
     **/
    Selectivity *		next;	
};

/** this constraint processor represents the main "iterator" of the
 *  matching function 
 **/
class SimpleCProcessor : public CProcessor {

public:
    /** initializes the constraint processor **/
    SimpleCProcessor(unsigned int ifcount, IProcessor & p, bitvector * v)
	: fmap(), if_mask(v), processor(p), target(ifcount) {};

    /** method called by the matching function for every matched
     *  constraint.
     *
     *  This method implements a fundamental part of the matching
     *  function.
     **/
    virtual bool		process_constraint(const Constraint *);

private:
    /** table of counters for partially-matched filters.
     *
     *  Each filter considered in the matching process has an
     *  associated counter that records the number of constraints
     *  matched so far for that filter
     **/
    map<const Filter *, unsigned short> fmap;    

    /** set of interfaces that can be ignored in the matching process.
     *
     *  An interface can be ignored by the matching process if it has
     *  already been matched and processed or if it was excluded by
     *  the pre-processing function.
     **/
    bitvector *			if_mask;
    /** output processor
     **/
    IProcessor &		processor;
    /** total number of interfaces we can match.
     *
     *  we maintain this value so that we can stop the matching
     *  process immediately once we have matched all interfaces.
     **/
    const unsigned int		target;
};

/** Interface descriptor.
 *
 *  Stores the association between user-supplied interface identifier
 *  and interface identifier used internally by the forwarding table.
 *  The reason for maintaining two identifiers is that identifiers
 *  used by the matching and pre-processing functions must be
 *  contiguous.  And we don't want to rely on the user to supply
 *  contiguous identifiers.
 **/
class Interface {
public:
    /** user-defined interface identifier **/
    const if_t			interface;

    /** internal identifier used by the matching function **/
    const ifid_t		id;

    /** builds an interface descriptor with the given identifiers **/
    Interface(if_t xif, ifid_t xid) : interface(xif), id(xid) {};
};

FwdTable::FwdTable(unsigned int blksize) 
    : memory(blksize), ifcount(0), preprocess_rounds(10), attributes(), 
      selectivity_first(NULL), selectivity_last(NULL) { }

void FwdTable::clear() {
    memory.clear();
    attributes.clear();
    ifcount = 0;
    selectivity_first = NULL;
    selectivity_last = NULL;
}

unsigned int FwdTable::bytesize() const {
    return memory.size();
}

Selectivity * FwdTable::new_selectivity(const string & n, ifid_t i) {
    char * s = new (memory) char[n.size() + 1];
    selectivity_last = new (memory) Selectivity(s, i, selectivity_last);

    for(string::const_iterator si = n.begin(); si != n.end(); ++si)
	*s++ = *si;		// copies the attribute name 
    *s = '\0';			// in the new buffer

    selectivity_last->exclude.set(i, memory);

    if (selectivity_first == NULL) 
	selectivity_first = selectivity_last;
    return selectivity_last;
}

void FwdTable::add_to_selectivity(Selectivity * s, ifid_t i) {
    s->exclude.set(i, memory);	// this increases the selectivity level by 1
    Selectivity * sp = s->prev;
    if (sp == NULL) return;

    while (sp->exclude.get_count() < s->exclude.get_count()) {
				// as long as the predecessor of s has
	sp->next = s->next;	// a lower selectivity level,
	s->next = sp;		// swap s and its predecessor
	if (sp->next == NULL) 
	    selectivity_last = sp;

	s->prev = sp->prev;
	sp->prev = s;
	if (s->prev == NULL) {
	    selectivity_first = s;
	    return;
	}
	sp = s->prev;
    }
}

Attribute * FwdTable::get_attribute(const sx_name & name) {
    TSTNode * np = attributes.insert(name.c_str(), memory);
    if (np->data == NULL) 
	np->data = new (memory) Attribute();

    return static_cast<Attribute*>(np->data);
}

Constraint * Attribute::add_constraint(const sx_constraint & c,
				       FTAllocator & a) {
    switch (c.value.type()) {
    case SX_int: 
	switch (c.op) {
	case SX_eq: return int_index.add_eq(c.value.int_value(), a); 
	case SX_gt: return int_index.add_gt(c.value.int_value(), a); 
	case SX_lt: return int_index.add_lt(c.value.int_value(), a); 
	default:
	    cerr << "bad operator for int." << endl;
	}
	break;
    case SX_string: 
	switch (c.op) {
	case SX_eq: return str_index.add_eq(c.value.string_value(), a);
	case SX_gt: return str_index.add_gt(c.value.string_value(), a);
	case SX_lt: return str_index.add_lt(c.value.string_value(), a);
	case SX_pf: return str_index.add_pf(c.value.string_value(), a);
	case SX_sf: return str_index.add_sf(c.value.string_value(), a);
	case SX_ss: return str_index.add_ss(c.value.string_value(), a);
	default:
	    cerr << "bad operator for string." << endl;
	}
	break;
    case SX_bool: 
	switch (c.op) {
	case SX_eq: return bool_index.add_eq(c.value.int_value(), a); 
	default:
	    cerr << "bad operator for boolean." << endl;
	}
	break;
    case SX_double: 
	switch (c.op) {
	case SX_eq: return double_index.add_eq(c.value.double_value(), a); 
	case SX_gt: return double_index.add_gt(c.value.double_value(), a); 
	case SX_lt: return double_index.add_lt(c.value.double_value(), a); 
	default:
	    cerr << "bad operator for double." << endl;
	}
	break;
    default:
	cerr << "bad type." << endl;
    }
    return NULL;
}

void FwdTable::connect(Constraint * c, Filter * f) {
    flist * n = new (memory)flist(f, c->filters);
    c->filters = n;
    ++(f->size);
}

//
//  This is the main method that builds the forwarding table.  It is
//  conceptually simple, but it requires a bit of attention to manage
//  the compilation of the selectivity table.
//
void FwdTable::ifconfig(if_t ifx, const sx_predicate & p) {
    sx_predicate::const_iterator	fi = p.begin();
    if (fi == p.end()) return;

    sx_filter::const_iterator	ci;
    Interface *		i;
    Filter *		f;
    Constraint *	c;
				
    typedef list<pair<string, Attribute *> > s_table;

    s_table		d_attrs;// slective attributes for this interface 
    s_table::iterator	di;	
    Attribute *		adescr;

    i = new (memory) Interface(ifx, ifcount++);
    f = NULL;
    //
    // now *fi is the first filter of predicate p, therefore we simply
    // fill d_attrs with every constraint name in *fi
    //
    for(ci = (*fi).begin(); ci != (*fi).end(); ++ci) {
	adescr = get_attribute((*ci).name);
	d_attrs.push_back(s_table::value_type((*ci).name, adescr));

	c = adescr->add_constraint(*ci, memory);
	if (c != NULL) {        // here we build the forwarding table
	    if (f == NULL)	// connecting constraints to the filter 
		f = new (memory) Filter(i);
	    connect(c, f);
	}			
    }				// then for every following filter *fi
    while(++fi != p.end()) {	// in p we intersect d_attrs with the
	f = NULL;		// sets of attributes of each of *fi
	di = d_attrs.begin();	
	for(ci = (*fi).begin(); ci != (*fi).end(); ++ci) {
				// we can do this because we know that
				// constraints come in alphabetical order
	    while(di != d_attrs.end() && (*di).first < (*ci).name) 
		d_attrs.erase(di++);
	    if (di != d_attrs.end() && (*di).first == (*ci).name) {
		c = (*di).second->add_constraint(*ci, memory);
		++di;
	    } else {
		c = get_attribute((*ci).name)->add_constraint(*ci, memory);
	    }
	    if (c != NULL) { 
		if (f == NULL)
		    f = new (memory) Filter(i);
		connect(c, f);
	    }
	}
	d_attrs.erase(di, d_attrs.end());
    }				// we add this interface to the exclude
				// set of each of the selective attributes
    for(di = d_attrs.begin(); di != d_attrs.end(); ++di) {
	if ((*di).second->exclude == NULL) {
	    (*di).second->exclude = new_selectivity((*di).first, i->id);
	} else {
	    add_to_selectivity((*di).second->exclude, i->id);
	}
    }
}

void FwdTable::preprocess_message(const sx_message & m, 
				  bitvector * v, unsigned int c) const {
    sx_message::const_iterator i;
    const Selectivity * si = selectivity_first;
    unsigned int count = 0; 
    while(count++ < preprocess_rounds && si != NULL && v->get_count() < c) {
	if ((i = m.find(si->name)) == m.end()) 
	    v->set(si->exclude);
	si = si->next;
    }
}

void FwdTable::set_preprocess_rounds(unsigned int i) {
    preprocess_rounds = i;
}

unsigned int FwdTable::get_preprocess_rounds() const {
    return preprocess_rounds;
}

void FwdTable::match(const sx_message & n, IProcessor & p) const {
    const TSTNode * tn;
    const Attribute * adescr;
    //
    // first we construct a constraint matcher object
    //
    bitvector mask(ifcount);
    preprocess_message(n, &mask, ifcount);
    SimpleCProcessor cp(ifcount, p, &mask);
    //
    //  here we go, the outer loop extracts the set of matching
    //  constraints for each attribute in n
    //
    for(sx_message::const_iterator i = n.begin(); i != n.end(); ++i) {
	if (mask.get_count() >= ifcount) return;
	tn = attributes.find((*i).first.c_str());
	if (tn != NULL) {
	    adescr = static_cast<Attribute *>(tn->data);
	    switch((*i).second.type()) {
	    case SX_int: 
		if (adescr->int_index.match((*i).second.int_value(), cp)) 
		    return;
		if (adescr->double_index.match((*i).second.int_value(), cp)) 
		    return;
		break;
	    case SX_string: 
		if (adescr->str_index.match((*i).second.string_value(), cp)) 
		    return;
		break;
	    case SX_bool: 
		if (adescr->bool_index.match((*i).second.bool_value(), cp)) 
		    return;
		break;
	    case SX_double: 
		if (adescr->double_index.match((*i).second.double_value(), cp)) 
		    return;
		if (adescr->int_index.match((int)(*i).second.double_value(), cp)) 
		    return;
		break;
	    default:
		cerr << "don't know how to deal with this type in notification!" << endl;
	    }
	}
    }
}

bool SimpleCProcessor::process_constraint(const Constraint * c) {
    //
    // we look at every filters in which the constraint appears
    //
    for(const flist * k = c->filters; k != NULL; k = k->next) {
	const Filter * f = k->f;
	if (! if_mask->test(f->i->id)) {
	    //
	    // if the interface to which the filter is
	    // associated has not been matched already
	    //
	    if (++fmap[f] >= f->size) {
		//
		// if the filter has been matched, then mark this
		// interface as "matched", and call the processor on
		// it.  interface id into the result.
		//
		if_mask->set(f->i->id);		
		if (processor.output(f->i->interface))
		    return true;
		if (if_mask->get_count() >= target) 
		    return true;
		//
		// we immediately return true if the processing
		// function returns true or if we matched all
		// possible interfaces
		//
	    }
	    //
	    // here's where I could do something intelligent
	    // about other filters pointing to the same
	    // interface: (*k)->interface->filters
	    //
	}
    }
    return false;
}
